import re
import geometry

class Token:
    text = ""
    indent = 0

class SOpen(Token):
    text = "("
    indent = 2

class SClose(Token):
    text = ")"
    indent = -2

class S:
    no_quoting = re.compile(r'^[a-z0-9\.\*:_-]+$', re.I)

    def __init__(self, tag, *params):
        self.tag = tag
        self.params = list(params)

    def append(self, *contents):
        self.params += list(contents)

    def __iter__(self):
        yield SOpen()
        yield self.tag

        for p in self.params:
            if isinstance(p, S):
                for pp in p:
                    yield pp
            else:
                yield p

        yield SClose()

    def code(self):
        indent = 0
        line_indent = 0
        ret = ""
        line = ""
        last = SOpen()

        for statement in self:
            spacing = " "
            if isinstance(statement, Token):
                text = statement.text
            else:
                if isinstance(statement, float):
                    text = ("%0.6f" % statement).rstrip("0")
                else:
                    text = str(statement)
                    if not self.no_quoting.match(text):
                        text = '"' + repr(text)[1:-1] + '"'

            if isinstance(last, SOpen) or isinstance(statement, SClose):
                spacing = ""

            if line and len(line) + len(spacing) + len(text) + line_indent > 80:
                if ret:
                    ret += "\n"
                if isinstance(last, SOpen):
                    text = line[-1] + text
                    line = line[:-1].strip()
                ret += " " * line_indent + line
                line = text
                line_indent = indent
            else:
                line += spacing + text

            if isinstance(statement, Token):
                indent += statement.indent

            last = statement

        if line:
            ret += "\n" + " " * indent + line
            line = ""

        return ret

class Point(geometry.Point):
    def s(self, tag = "at"):
        return S(tag, self.x, self.y)

class Rectangle(geometry.Rectangle):
    point_class = Point

class Module:
    def __init__(self, name, descr, mode = "smd"):
        self.name = name
        self.descr = descr
        self.mode = mode
        self.contents = []
        self.bbox = Rectangle(0, 0, 0, 0)

    def append(self, *objs):
        for o in objs:
            self.bbox |= o.bbox()
            self.contents.append(o)

    def s(self):
        root = S("module", self.name,
                 S("layer", "F.Cu"),
                 S("descr", self.descr),
                 S("attr", self.mode),
                 Reference(self.bbox.tc - Point(0, Font.default.height * .6)).s(),
                 Value(self.bbox.bc + Point(0, Font.default.height * .6),
                      self.descr).s(),
                 Line(self.bbox.tl, self.bbox.tr, .05, "F.Fab").s(),
                 Line(self.bbox.tr, self.bbox.br, .05, "F.Fab").s(),
                 Line(self.bbox.br, self.bbox.bl, .05, "F.Fab").s(),
                 Line(self.bbox.bl, self.bbox.tl, .05, "F.Fab").s(),
        )

        for c in self.contents:
            assert isinstance(c, Grob), ValueError(c)
            root.append(c.s())

        return root

    def code(self):
        return self.s().code()

class Grob:
    def __init__(self, layer):
        self.layer = layer

    def bbox(self):
        return Rectangle(0, 0, 0, 0)

class Font:
    def __init__(self, width, height, thickness = .15, variant = "normal"):
        self.width = width
        self.height = height
        self.thickness = thickness
        self.variant = variant

    def s(self):
        return S("font",
                 S("size", self.width, self.height),
                 S("thickness", self.thickness))

Font.default = Font(.7, .7, .15)
Font.small = Font(.5, .5, .1)

class Info(Grob):
    def __init__(self, position, type, text, layer, font = Font.default):
        Grob.__init__(self, layer)
        self.position = position
        self.text = text
        self.type = type
        self.font = font

    def s(self):
        return S("fp_text", self.type, self.text,
                 self.position.s(),
                 S("layer", self.layer),
                 S("effects", self.font.s()))

class Reference(Info):
    def __init__(self, position):
        Info.__init__(self, position, "reference", "REF**", "F.SilkS")

class Value(Info):
    def __init__(self, position, text):
        Info.__init__(self, position, "value", text, "F.Fab")

class Text(Info):
    def __init__(self, position, text, layer = "F.Fab", font = Font.default):
        Info.__init__(self, position, "user", text, layer, font)

class Line(Grob):
    def __init__(self, end1, end2, width, layer):
        Grob.__init__(self, layer)
        self.ends = (end1, end2)
        self.width = width

    def s(self):
        return S("fp_line", self.ends[0].s("start"), self.ends[1].s("end"),
                 S("layer", self.layer),
                 S("width", self.width))

    def bbox(self):
        return Rectangle.containing(*self.ends).enlarged(self.width)

class Via(Grob):
    def __init__(self, position, hole_diameter, ring_diameter):
        Grob.__init__(self, "*.Cu")
        self.position = position
        self.hole_diameter = hole_diameter
        self.ring_diameter = ring_diameter

    def s(self):
        return S("pad", "", "thru_hole", "circle",
                 self.position.s(),
                 S("size", self.ring_diameter, self.ring_diameter),
                 S("drill", self.hole_diameter),
                 S("layers", self.layer))

    def bbox(self):
        return Rectangle(self.position.y - self.ring_diameter / 2,
                                  self.position.x + self.ring_diameter / 2,
                                  self.position.y + self.ring_diameter / 2,
                                  self.position.x - self.ring_diameter / 2)

class Strip(Grob):
    def __init__(self, center, size, layer = "F.Cu", angle = 0):
        Grob.__init__(self, layer)
        self.center = center
        self.size = size
        self.angle = angle

    def s(self):
        pos = self.center.s()
        if self.angle:
            pos.append(self.angle)

        return S("pad", "", "smd", "rect",
                 pos, self.size.s("size"),
                 S("layers", self.layer))

    def bbox(self):
        return Rectangle(self.center.y - self.size.y / 2,
                                  self.center.x + self.size.x / 2,
                                  self.center.y + self.size.y / 2,
                                  self.center.x - self.size.x / 2)

class SmdSquarePad(Strip):
    def __init__(self, center, size, number = None, layer = "F.Cu", angle = 0):
        Grob.__init__(self, layer)
        self.center = center
        self.size = size
        self.number = "" if number is None else number
        self.angle = angle

    def s(self):
        pos = self.center.s()
        if self.angle:
            pos.append(self.angle)

        if self.layer in ("F.Cu", "B.Cu"):
            layers = self.layer, self.layer[0] + ".Paste", self.layer[0] + ".Mask"
        else:
            layers = self.layer,

        return S("pad", self.number, "smd", "rect",
                 pos, self.size.s("size"),
                 S("layers", *layers))

class SmdRoundPad(Grob):
    def __init__(self, position, diameter, number = None, layer = "F.Cu"):
        Grob.__init__(self, layer)
        self.position = position
        self.diameter = diameter
        self.number = "" if number is None else number

    def s(self):
        return S("pad", self.number, "smd", "rect",
                 self.position.s(),
                 S("size", self.diameter, self.diameter),
                 S("layers", "F.Cu", "F.Paste", "F.Mask"))

    def bbox(self):
        return Rectangle(self.position.y - self.diameter / 2,
                                  self.position.x + self.diameter / 2,
                                  self.position.y + self.diameter / 2,
                                  self.position.x - self.diameter / 2)

class Connection(Grob):
    def __init__(self, position, diameter, number, layer = "F.Cu"):
        Grob.__init__(self, layer)
        self.position = position
        self.diameter = diameter
        self.number = number

    def s(self):
        return S("pad", self.number, "connect", "circle",
                 self.position.s(),
                 S("size", self.diameter, self.diameter),
                 S("layers", self.layer))

    def bbox(self):
        return Rectangle(self.position.y - self.diameter / 2,
                                  self.position.x + self.diameter / 2,
                                  self.position.y + self.diameter / 2,
                                  self.position.x - self.diameter / 2)

def test():
    m = Module("TI_DN007", "2.4GHz antenna", "virtual")
    for (x, y, sx, sy, a) in [
            (0, -2.545, 25.58, 1.21, 0),
            (-7.71, -1.44, 5.8, 1, 0),
            (-11.7, -0.065, 2.18, 6.17, 0),
            (-7.7, 2.375, 10.18, 1.29, 0),
            (-4.21, 0.465, 3.2, 1.21, 0),
            (-4.21, -1.04, 1.2, 4.22, 0),
            (-2.332, 1.046, 0.46, 1.326, 45),
            (-5.235, 2.745, 1.15, 2.03, 0),
            (-3.185, 2.795, 1.15, 1.93, 0),
            (-1.93, 2.556, 0.46, 2.408, 0),
    ]:
        m.append(Strip(Point(x, y), Point(sx, sy), angle = a))
    m.append(Connection(Point(-5.235, 3.65), .3, 1))
    m.append(Connection(Point(-3.185, 3.65), .3, 1))
    m.append(Connection(Point(-1.93, 3.65), .3, 2))

    print m.code()

def Ti_DN007():
    H1 = 5.70
    H2 = 0.74
    H3 = 1.29
    H4 = 2.21
    H5 = 0.66
    H6 = 1.21
    H7 = 0.80
    H8 = 1.80
    H9 = 0.61
    W1 = 1.21
    W2 = 0.46
    L1 = 25.58
    L2 = 16.40
    L3 = 2.18
    L4 = 4.80
    L5 = 1.00
    L6 = 1.00
    L7 = 3.20
    L8 = 0.45

    m = Module("TI_DN007", "2.4GHz antenna", "virtual")
    for (x, y, w, h) in [
            (0, 0, L1, W1),
            (0, 0, L3 + L4 + L7 - L6, H4),
            (0, 0, L3, H3 + H4 + H5 + H6),
            (0, H4 + H5 + H6, L3 + L4 + L7, H3),
            (L3 + L4, H4 + H5 + H6 + H7, L5, H2),
            (L3 + L4 + L7 - L6, H4 + H5 + H6 + H7, L6, H2),
            (L3 + L4 + L5, W1, L7 - L6 - L5, H8),
            (L3 + L4, W1 + H8, L7, H6),
    ]:
        m.append(Strip(Point(x, y), Point(x, h)))
    print m.code()

if __name__ == "__main__":
    test()
#    Ti_DN007()
